import signal
import sys
from datetime import datetime
from pathlib import Path
from pathlib import PurePosixPath
from types import FrameType
from typing import TypedDict

import click
from shepherd_core.data_models.task import EmulationTask
from shepherd_core.data_models.task import HarvestTask
from shepherd_core.data_models.task import ProgrammingTask
from shepherd_core.data_models.testbed import ProgrammerProtocol
from shepherd_core.data_models.testbed import TargetPort
from typing_extensions import Unpack

from . import __version__
from .herd import Herd
from .logger import activate_verbosity
from .logger import log

# TODO:
#  - click.command shorthelp can also just be the first sentence of docstring
#  https://click.palletsprojects.com/en/8.1.x/documentation/#command-short-help
#  - document arguments in their docstring (has no help=)
#  - arguments can be configured in a dict and standardized across tools


def exit_gracefully(_signum: int, _frame: FrameType | None) -> None:
    """Signal-handling for a clean exit-strategy."""
    log.warning("Exiting!")
    sys.exit(0)


@click.group(context_settings={"help_option_names": ["-h", "--help"], "obj": {}})
@click.option(
    "--inventory",
    "-i",
    type=click.STRING,
    default=None,
    help="List of target hosts as comma-separated string or path to ansible-style yaml file",
)
@click.option(
    "--limit",
    "-l",
    type=click.STRING,
    default=None,
    help="Comma-separated list of hosts to limit execution to",
)
@click.option(
    "--user",
    "-u",
    type=click.STRING,
    default=None,
    help="User name for login to nodes",
)
@click.option(
    "--key-filepath",
    "-k",
    type=click.Path(exists=True, readable=True, file_okay=True, dir_okay=False, path_type=Path),
    default=None,
    help="Path to private ssh key file",
)
@click.option("--verbose", "-v", is_flag=True)
@click.pass_context
def cli(
    ctx: click.Context,
    inventory: str | None,
    limit: str | None,
    user: str | None,
    key_filepath: Path | None,
    *,
    verbose: bool,
) -> None:
    """Entry for command line with settings to interface the herd."""
    signal.signal(signal.SIGTERM, exit_gracefully)
    signal.signal(signal.SIGINT, exit_gracefully)

    if verbose:
        activate_verbosity()

    if not ctx.invoked_subcommand:
        click.echo("Please specify a valid command")

    ctx.obj["herd"] = Herd(inventory, limit, user, key_filepath)


# #############################################################################
#                               Misc-Commands
# #############################################################################


@cli.command(short_help="Print version-info (combine with -v for more)")
def version() -> None:
    """Print version-info (combine with -v for more)."""
    from h5py import __version__ as ver_h5py
    from numpy import __version__ as ver_numpy
    from pydantic import __version__ as ver_pydantic
    from shepherd_core import __version__ as ver_core
    from yaml import __version__ as ver_yaml

    log.info("Shepherd-Herd v%s", __version__)
    log.info("Shepherd-core v%s", ver_core)
    log.debug("Python v%s", sys.version)
    log.debug("click v%s", click.__version__)
    log.debug("h5py v%s", ver_h5py)
    log.debug("numpy v%s", ver_numpy)
    log.debug("pydantic v%s", ver_pydantic)
    log.debug("PyYAML v%s", ver_yaml)


@cli.command(
    short_help="Power off shepherd observers."
    "Be sure to have physical access to the hardware "
    "for manually starting them again."
)
@click.option("--restart", "-r", is_flag=True, help="Reboot")
@click.pass_context
def poweroff(ctx: click.Context, *, restart: bool) -> None:
    """Power off shepherd observers."""
    with ctx.obj["herd"] as herd:
        exit_code = herd.poweroff(restart=restart)
    ctx.exit(exit_code)


@cli.command(short_help="Run COMMAND on the shell")
@click.pass_context
@click.argument("command", type=click.STRING)
@click.option("--sudo", "-s", is_flag=True, help="Run command with sudo")
def shell_cmd(ctx: click.Context, command: str, *, sudo: bool) -> None:
    """Run COMMAND on the shell."""
    with ctx.obj["herd"] as herd:
        replies = herd.run_cmd(sudo=sudo, cmd=command)
        herd.print_output(replies, verbose=True)
        exit_code = max([0] + [abs(reply.exited) for reply in replies.values()])
    ctx.exit(exit_code)


@cli.command(short_help="Collects information about the observer-hosts -> saved to local file")
@click.argument(
    "output-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    default=Path("./"),
)
@click.pass_context
def inventorize(ctx: click.Context, output_path: Path) -> None:
    """Collect information about the observer-hosts -> saved to local file."""
    with ctx.obj["herd"] as herd:
        failed = herd.inventorize(output_path)
    ctx.exit(int(failed))


@cli.command(
    short_help="Reloads the shepherd-kernel-module on each sheep",
    context_settings={"ignore_unknown_options": True},
)
@click.pass_context
def fix(ctx: click.Context) -> None:
    """Reload the shepherd-kernel-module on each sheep."""
    with ctx.obj["herd"] as herd:
        replies = herd.run_cmd(
            sudo=True,
            cmd="shepherd-sheep fix",
        )
        herd.print_output(replies, verbose=False)
        exit_code = max([0] + [abs(reply.exited) for reply in replies.values()])
    ctx.exit(exit_code)


@cli.command(
    short_help="Gets current time and restarts PTP on each sheep",
    context_settings={"ignore_unknown_options": True},
)
@click.pass_context
def resync(ctx: click.Context) -> None:
    """Get current time and restarts PTP on each sheep."""
    activate_verbosity()
    with ctx.obj["herd"] as herd:
        exit_code = herd.resync()
    ctx.exit(exit_code)


@cli.command(
    short_help="Helps to identify Observers by flashing LEDs near Targets (IO, EMU)",
    context_settings={"ignore_unknown_options": True},
)
@click.argument("duration", type=click.INT, default=30)
@click.pass_context
def blink(ctx: click.Context, duration: int) -> None:
    """Help to identify Observers by flashing LEDs near Targets (IO, EMU)."""
    with ctx.obj["herd"] as herd:
        replies = herd.run_cmd(
            sudo=True,
            cmd=f"shepherd-sheep blink {duration}",
        )
        herd.print_output(replies, verbose=False)
        exit_code = max([0] + [abs(reply.exited) for reply in replies.values()])
    ctx.exit(exit_code)


@cli.command(
    short_help="Check if all remote hosts are present & responding.",
    context_settings={"ignore_unknown_options": True},
)
@click.pass_context
def alive(ctx: click.Context) -> None:
    with ctx.obj["herd"] as herd:
        failed = not herd.alive()
    if failed:
        log.warning("Not all remote hosts are responding.")
    else:
        log.debug("All remote hosts are responding.")
    ctx.exit(int(failed))


# #############################################################################
#                               Task-Handling
# #############################################################################


@cli.command(
    short_help="Run a task or set of tasks with provided config/task file (YAML). "
    "NOTE: if no start-time is present, observers will not start synchronously.",
)
@click.argument(
    "config",
    type=click.Path(exists=True, readable=True, file_okay=True, dir_okay=False, path_type=Path),
)
@click.option("--attach", "-a", is_flag=True, help="Wait and receive output on shell")
@click.pass_context
def run(ctx: click.Context, config: Path, *, attach: bool) -> None:
    """Run a task or set of tasks with provided config/task file (YAML)."""
    with ctx.obj["herd"] as herd:
        exit_code = herd.run_task(config, attach=attach)
    ctx.exit(exit_code)


@cli.command(
    short_help="Simultaneously record IV data from the connected "
    "harvesting-sources on the chosen observers."
)
@click.option(
    "--output-path",
    "-o",
    type=click.Path(dir_okay=True, file_okay=True, path_type=PurePosixPath),
    default=Herd.path_default,
    help="Dir or file path for resulting hdf5 file",
)
@click.option(
    "--virtual-harvester",
    "-a",
    type=click.STRING,
    default=None,
    help="Choose one of the predefined virtual harvesters",
)
@click.option(
    "--duration",
    "-d",
    type=click.FLOAT,
    default=None,
    help="Duration of recording in seconds",
)
@click.option("--force-overwrite", "-f", is_flag=True, help="Overwrite existing file")
@click.option(
    "--use-cal-default",
    "-c",
    is_flag=True,
    help="Use default calibration values",
)
@click.option(
    "--no-start",
    "-n",
    is_flag=True,
    help="Start shepherd synchronized after uploading config",
)
@click.pass_context
def harvest(
    ctx: click.Context,
    *,
    no_start: bool,
    **kwargs: Unpack[TypedDict],
) -> None:
    """Simultaneously record IV data from harvesting-sources on the chosen observers."""
    with ctx.obj["herd"] as herd:
        for path in ["output_path"]:
            file_path = PurePosixPath(kwargs[path])
            if not file_path.is_absolute():
                kwargs[path] = Herd.path_default / file_path
            kwargs[path] = Path(kwargs[path])
            # ⤷ TODO: workaround until shepherd-core uses PurePaths

        if kwargs.get("virtual_harvester") is not None:
            kwargs["virtual_harvester"] = {"name": kwargs["virtual_harvester"]}

        ts_start = datetime.now().astimezone()
        delay = 0
        if not no_start:
            ts_start, delay = herd.find_consensus_time()
            kwargs["time_start"] = ts_start

        kwargs = {key: value for key, value in kwargs.items() if value is not None}
        task = HarvestTask(**kwargs)
        herd.put_task(task)

        if not no_start:
            log.info(
                "Scheduling start of shepherd: %s (in ~ %.2f s)",
                ts_start.isoformat(),
                delay,
            )
            exit_code = herd.start_measurement()
            log.info("Shepherd started.")
            if exit_code > 0:
                log.debug("-> max exit-code = %d", exit_code)


@cli.command(
    short_help="Use the previously recorded harvest-data "
    "(INPUT-PATH is a hdf5-file on the sheep-hosts) "
    "for emulating an energy environment for the attached "
    "sensor nodes and monitor their power consumption and GPIO events",
)
@click.argument(
    "input-path",
    type=click.Path(file_okay=True, dir_okay=False, path_type=PurePosixPath),
)
# TODO: switch to local file for input?
@click.option(
    "--output-path",
    "-o",
    type=click.Path(dir_okay=True, file_okay=True, path_type=PurePosixPath),
    default=Herd.path_default,
    help="Dir or file path for resulting hdf5 file with load recordings",
)
@click.option(
    "--duration",
    "-d",
    type=click.FLOAT,
    default=None,
    help="Duration of recording in seconds",
)
@click.option("--force-overwrite", "-f", is_flag=True, help="Overwrite existing file")
@click.option(
    "--use-cal-default",
    "-c",
    is_flag=True,
    help="Use default calibration values",
)
@click.option(
    "--enable-io/--disable-io",
    default=True,
    help="Switch the GPIO level converter to targets on/off",
)
@click.option(
    "--io-port",
    type=click.Choice(["A", "B"]),
    default="A",
    help="Choose Target that gets connected to IO",
)
@click.option(
    "--pwr-port",
    type=click.Choice(["A", "B"]),
    default="A",
    help="Choose (main)Target that gets connected to virtual Source / current-monitor",
)
@click.option(
    "--voltage-aux",
    "-x",
    type=click.FLOAT,
    default=0.0,
    help="Set Voltage of auxiliary Power Source (second target)",
)
@click.option(
    "--virtual-source",
    "-a",  # -v & -s already taken for sheep, so keep it consistent with hrv (algorithm)
    type=click.STRING,
    default=None,
    help="Use the desired setting for the virtual source",
)
@click.option(
    "--no-start",
    "-n",
    is_flag=True,
    help="Start shepherd synchronized after uploading config",
)
@click.pass_context
def emulate(
    ctx: click.Context,
    *,
    no_start: bool,
    **kwargs: Unpack[TypedDict],
) -> None:
    """Emulate an energy environment for the attached sensor nodes.

    Use the previously recorded harvest-data for emulating an energy environment for the attached
    sensor nodes and monitor their power consumption and GPIO events
    - INPUT-PATH is a hdf5-file on the sheep-hosts
    """
    with ctx.obj["herd"] as herd:
        for path in ["input_path", "output_path"]:
            file_path = PurePosixPath(kwargs[path])
            if not file_path.is_absolute():
                kwargs[path] = Herd.path_default / file_path
            kwargs[path] = Path(kwargs[path])
            # ⤷ TODO: workaround until shepherd-core uses PurePaths

        for port in ["io_port", "pwr_port"]:
            kwargs[port] = TargetPort[kwargs[port]]

        if kwargs.get("virtual_source") is not None:
            kwargs["virtual_source"] = {"name": kwargs["virtual_source"]}

        ts_start = datetime.now().astimezone()
        delay = 0
        if not no_start:
            ts_start, delay = herd.find_consensus_time()
            kwargs["time_start"] = ts_start

        kwargs = {key: value for key, value in kwargs.items() if value is not None}
        task = EmulationTask(**kwargs)
        herd.put_task(task)

        if not no_start:
            log.info(
                "Scheduling start of shepherd: %s (in ~ %.2f s)",
                ts_start.isoformat(),
                delay,
            )
            exit_code = herd.start_measurement()
            log.info("Shepherd started.")
            if exit_code > 0:
                log.debug("-> max exit-code = %d", exit_code)


# #############################################################################
#                               Controlling Measurements
# #############################################################################


@cli.command(
    short_help="Start pre-configured shp-service (/etc/shepherd/config.yml, "
    "UNSYNCED when 'time_start' is not set)",
)
@click.pass_context
def start(ctx: click.Context) -> None:
    """Start pre-configured shp-service.

    - source /etc/shepherd/config.yml,
    - UNSYNCED when 'time_start' is not set)
    """
    ret: int = 0
    with ctx.obj["herd"] as herd:
        if herd.check_status():
            log.info("Shepherd still active, will skip this command!")
            ret = 1
        else:
            exit_code = herd.start_measurement()
            log.info("Shepherd started.")
            if exit_code > 0:
                log.debug("-> max exit-code = %d", exit_code)
    ctx.exit(ret)


@cli.command(short_help="Information about current state of shepherd measurement")
@click.pass_context
def status(ctx: click.Context) -> None:
    """Information about current state of shepherd measurement."""
    ret: int = 0
    with ctx.obj["herd"] as herd:
        if herd.check_status():
            log.info("Shepherd still active!")
            ret = 1
        else:
            log.info("Shepherd not active! (measurement is done)")
        delta = herd.get_last_usage()
        if delta is not None:
            ts_now = datetime.now().astimezone()
            log.info("Last usage was %s, Δt = %s", str(ts_now - delta), str(delta))
    ctx.exit(ret)


@cli.command(short_help="Stops any harvest/emulation or other processes blocking the sheep")
@click.pass_context
def stop(ctx: click.Context) -> None:
    """Stop any harvest/emulation or other processes blocking the sheep."""
    with ctx.obj["herd"] as herd:
        exit_code = herd.stop_measurement()
    log.info("Shepherd stopped.")
    if exit_code > 0:
        log.debug("-> max exit-code = %d", exit_code)


# #############################################################################
#                               File Handling
# #############################################################################


@cli.command(
    short_help="Uploads a file FILENAME to the remote observers, will be stored in REMOTE_PATH",
)
@click.argument(
    "filename",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path),
)
@click.option(
    "--remote-path",
    "-r",
    type=click.Path(file_okay=True, dir_okay=True, path_type=PurePosixPath),
    default=Herd.path_default,
    help="for safety only allowed: /var/shepherd/* or /etc/shepherd/*",
)
@click.option("--force-overwrite", "-f", is_flag=True, help="Overwrite existing file")
@click.pass_context
def distribute(
    ctx: click.Context,
    filename: Path,
    remote_path: PurePosixPath,
    *,
    force_overwrite: bool,
) -> None:
    """Upload a file FILENAME to the remote observers, which will be stored in REMOTE_PATH."""
    with ctx.obj["herd"] as herd:
        herd.put_file(filename, remote_path, force_overwrite=force_overwrite)


@cli.command(short_help="Retrieves remote hdf file FILENAME and stores in OUTDIR")
@click.argument("filename", type=click.Path(file_okay=True, dir_okay=False, path_type=Path))
@click.argument(
    "outdir",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        path_type=Path,
    ),
)
@click.option(
    "--timestamp",
    "-t",
    is_flag=True,
    help="Add current timestamp to measurement file",
)
@click.option(
    "--separate",
    "-s",
    is_flag=True,
    help="Every remote node gets own subdirectory",
)
@click.option(
    "--delete",
    "-d",
    is_flag=True,
    help="Delete the file from the remote filesystem after retrieval",
)
@click.option(
    "--force-stop",
    "-f",
    is_flag=True,
    help="Stop the on-going harvest/emulation process before retrieving the data",
)
@click.pass_context
def retrieve(
    ctx: click.Context,
    filename: Path,
    outdir: Path,
    *,
    timestamp: bool,
    separate: bool,
    delete: bool,
    force_stop: bool,
) -> None:
    """Retrieve remote hdf file FILENAME and stores in OUTDIR.

    filename: can either be
    (a) remote file with absolute path or relative path in '/var/shepherd/recordings/' or
    (b) local job- / task-file that did already run (embedded paths are retrieved)

    outdir: local path to put the files in 'outdir/[node-name]/filename'
    """
    with ctx.obj["herd"] as herd:
        if force_stop:
            herd.stop_measurement()
            if herd.await_stop(timeout=30):
                raise TimeoutError("shepherd still active after timeout")

        if (
            filename.is_file()
            and filename.exists()
            and filename.suffix in [".yaml", ".yml", ".pickle"]
        ):
            failed = herd.get_task_files(filename, outdir, separate=separate, delete_src=delete)
        else:
            filename = PurePosixPath(filename)
            failed = herd.get_file(
                filename,
                outdir,
                timestamp=timestamp,
                separate=separate,
                delete_src=delete,
            )
    ctx.exit(int(failed))


# #############################################################################
#                               Pru Programmer
# #############################################################################


@cli.command(
    short_help="Programmer for Target-Controller",
)
@click.argument(
    "firmware-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path),
)
@click.option(
    "--target-port",
    "-p",
    type=click.Choice(["A", "B"]),
    default="A",
    help="Choose Target-Port of Cape for programming",
)
@click.option(
    "--mcu-port",
    "-m",
    type=click.INT,
    default=1,
    help="Choose MCU on Target-Port (only valid for SBW & SWD)",
)
@click.option(
    "--voltage",
    "-v",
    type=click.FLOAT,
    default=None,
    help="Target supply voltage",
)
@click.option(
    "--datarate",
    "-d",
    type=click.INT,
    default=None,
    help="Bit rate of Programmer (bit/s)",
)
@click.option(
    "--mcu-type",
    "-t",
    type=click.Choice(["nrf52", "msp430"]),
    default="nrf52",
    help="Target MCU",
)
@click.option(
    "--simulate",
    is_flag=True,
    help="dry-run the programmer - no data gets written",
)
@click.pass_context
def program(ctx: click.Context, **kwargs: Unpack[TypedDict]) -> None:
    """Programmer for Target-Controller."""
    tmp_file = PurePosixPath("/tmp/target_image.hex")  # noqa: S108
    cfg_path = PurePosixPath("/etc/shepherd/config_for_herd.pickle")

    with ctx.obj["herd"] as herd:
        herd.put_file(kwargs["firmware_file"], tmp_file, force_overwrite=True)
        protocol_dict = {
            "nrf52": ProgrammerProtocol.swd,
            "msp430": ProgrammerProtocol.sbw,
        }
        kwargs["protocol"] = protocol_dict[kwargs["mcu_type"]]
        kwargs["firmware_file"] = PurePosixPath(tmp_file)
        kwargs["firmware_file"] = Path(kwargs["firmware_file"])
        # ⤷ TODO: workaround until shepherd-core uses PurePaths

        kwargs = {key: value for key, value in kwargs.items() if value is not None}
        task = ProgrammingTask(**kwargs)
        herd.put_task(task, cfg_path)

        command = f"shepherd-sheep --verbose run {cfg_path.as_posix()}"
        replies = herd.run_cmd(sudo=True, cmd=command)
        exit_code = max([0] + [abs(reply.exited) for reply in replies.values()])
        if exit_code:
            log.error("Programming - Procedure failed - will exit now!")
        herd.print_output(replies, verbose=False)
    ctx.exit(exit_code)


if __name__ == "__main__":
    cli()



import time

import pytest
from click.testing import CliRunner
from shepherd_herd.herd_cli import cli

from .conftest import wait_for_end


@pytest.mark.timeout(30)
@pytest.mark.usefixtures("_herd_alive")
def test_herd_alive() -> None:
    pass


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_hrv_example(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "harvest",
            "-a",
            "cv20",
            "-d",
            "10",
            "-o",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_stopped")
def test_hrv_example_fail(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "harvest",
            "--virtual-harvester",
            "ceeeveeeee",
            "--duration",
            "10",
            "--output-path",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code != 0
    wait_for_end(cli_runner, timeout=15)


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_stopped")
def test_hrv_minimal(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        ["harvest"],
    )
    assert res.exit_code == 0
    time.sleep(10)
    # forced stop
    res = cli_runner.invoke(
        cli,
        ["-v", "stop"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=10)


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_hrv_all_args_long(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "harvest",
            "--virtual-harvester",
            "cv33",
            "--duration",
            "10",
            "--force-overwrite",
            "--use-cal-default",
            "--output-path",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_hrv_all_args_short(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "harvest",
            "-a",
            "cv33",
            "-d",
            "10",
            "-f",
            "-c",
            "-o",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_hrv_no_start(cli_runner: CliRunner) -> None:
    # Note: short timeout is the catch
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "harvest",
            "-d",
            "10",
            "--no-start",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=15)
    # manual start
    res = cli_runner.invoke(
        cli,
        ["-v", "start"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


# TODO: retrieve & verify with shepherd-core (length & validity)



import time
from pathlib import Path
from shutil import copy

import numpy as np
import pytest
import yaml
from click.testing import CliRunner
from shepherd_data import Writer
from shepherd_herd import Herd
from shepherd_herd.herd import path_xdg_config
from shepherd_herd.herd_cli import cli


@pytest.fixture
def cli_runner() -> CliRunner:
    return CliRunner()


def extract_first_sheep(herd_path: Path) -> str:
    with herd_path.open(encoding="utf-8-sig") as stream:
        try:
            inventory_data = yaml.safe_load(stream)
        except yaml.YAMLError as _xpt:
            msg = f"Couldn't read inventory file {herd_path.as_posix()}"
            raise TypeError(msg) from _xpt
    return next(iter(inventory_data["sheep"]["hosts"].keys()))


def wait_for_end(cli_runner: CliRunner, tmin: float = 0, timeout: float = 999) -> bool:
    ts_start = time.time()
    while cli_runner.invoke(cli, ["status"]).exit_code > 0:
        duration = time.time() - ts_start
        if duration > timeout:
            msg = f"Shepherd ran into timeout ({timeout} s)"
            raise TimeoutError(msg)
        time.sleep(2)
    duration = time.time() - ts_start
    if duration < tmin:
        msg = f"Shepherd only took {duration} s (min = {tmin} s)"
        raise TimeoutError(msg)
    return False


def generate_h5_file(file_path: Path, file_name: str = "harvest_example.h5") -> Path:
    store_path = file_path / file_name

    with Writer(store_path, compression=None) as file:
        file.store_hostname("artificial")
        duration_s = 2
        repetitions = 10
        timestamp_vector = np.arange(0.0, duration_s, file.sample_interval_ns / 1e9)

        # values in SI units
        voltages = np.linspace(3.30, 1.90, int(file.samplerate_sps * duration_s))
        currents = np.linspace(100e-6, 2000e-6, int(file.samplerate_sps * duration_s))

        for idx in range(repetitions):
            timestamps = idx * duration_s + timestamp_vector
            file.append_iv_data_si(timestamps, voltages, currents)

    return store_path


@pytest.fixture
def data_h5_path(tmp_path: Path) -> Path:
    return generate_h5_file(tmp_path)


@pytest.fixture
def local_herd(tmp_path: Path) -> Path:
    # locations copied from herd.__innit__()
    inventories: list[Path] = [
        Path().cwd() / "herd.yaml",
        Path().cwd() / "inventory/herd.yaml",
        Path("~").expanduser() / "herd.yaml",
        path_xdg_config / "shepherd/herd.yaml",
        Path("/etc/shepherd/herd.yaml"),
    ]
    host_path = None
    for inventory in inventories:
        if Path(inventory).exists():
            host_path = Path(inventory)
    if host_path is None:
        raise FileNotFoundError(", ".join(inventories))
    local_path = tmp_path / "herd.yaml"
    copy(host_path, local_path)

    return local_path


@pytest.fixture
def _herd_alive() -> None:
    # pre-test is good for start of new test-file
    for _ in range(3):
        time.sleep(1)
        with Herd() as herd:
            if len(herd.group) > 0:
                time.sleep(1)
                return
    raise RuntimeError("No Sheep seems to be alive")


@pytest.fixture
def _herd_stopped(cli_runner: CliRunner, _herd_alive: None) -> None:
    cli_runner.invoke(cli, ["-v", "stop"])
    wait_for_end(cli_runner)
    # make sure kernel module is active
    cli_runner.invoke(
        cli,
        [
            "-v",
            "fix",
        ],
    )



import time
from pathlib import Path

import pytest
from click.testing import CliRunner
from shepherd_herd.herd_cli import cli

from .conftest import generate_h5_file
from .conftest import wait_for_end


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_prepare(cli_runner: CliRunner, tmp_path: Path) -> None:
    # distribute file and emulate from it in following tests
    test_file = generate_h5_file(tmp_path, "pytest_src.h5")
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "distribute",
            "--force-overwrite",
            test_file.as_posix(),
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner)


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_example(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "emulate",
            "--virtual-source",
            "BQ25504",
            "-o",
            "pytest_emu.h5",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=20)


@pytest.mark.timeout(80)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_example_fail(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "emulate",
            "--virtual-source",
            "BQ25504",
            "-o",
            "pytest_emu.h5",
            "pytest_NonExisting.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=40)  # TODO: was 15 but got worse with core-lib


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_minimal(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "emulate",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=20)


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_all_args_long(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "emulate",
            "--duration",
            "10",
            "--force-overwrite",
            "--use-cal-default",
            "--enable-io",
            "--io-port",
            "A",
            "--pwr-port",
            "A",
            "--voltage-aux",
            "1.6",
            "--virtual-source",
            "BQ25504",
            "--output-path",
            "pytest_emu.h5",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_all_args_short(cli_runner: CliRunner) -> None:
    # short arg or opposite bool val
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "emulate",
            "-d",
            "10",
            "-f",
            "-c",
            "--disable-io",
            "--io-port",
            "B",
            "--pwr-port",
            "B",
            "-x",
            "1.4",
            "-a",
            "BQ25570",
            "-o",
            "pytest_emu.h5",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(150)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_no_start(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "emulate",
            "-d",
            "20",
            "-o",
            "pytest_emu.h5",
            "--no-start",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=15)
    # manual start
    res = cli_runner.invoke(
        cli,
        ["-v", "start"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_stopped")
def test_emu_force_stop(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "emulate",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    time.sleep(10)
    # forced stop
    res = cli_runner.invoke(
        cli,
        ["-v", "stop"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=10)


# TODO: retrieve & verify with shepherd-core



"""shepherd_herd command line utility.

click-based command line utility for controlling a group of shepherd nodes
remotely through ssh. Provides commands for starting/stopping harvester and
emulator, retrieving recordings to the local machine and flashing firmware
images to target sensor nodes.

"""

from .herd import Herd
from .logger import activate_verbosity
from .logger import get_verbosity
from .logger import log

__version__ = "0.9.3"

__all__ = [
    "Herd",
    "activate_verbosity",
    "get_verbosity",
    "log",
    "log",
]



"""Herd-Baseclass for controlling the sheep-observer-nodes."""

import contextlib
import logging
import os
import pickle
import threading
import time
from collections.abc import Mapping
from collections.abc import Set as AbstractSet
from datetime import datetime
from datetime import timedelta
from io import BytesIO
from pathlib import Path
from pathlib import PurePath
from pathlib import PurePosixPath
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Any

import yaml
from fabric import Connection
from fabric import Group
from fabric import Result
from paramiko.ssh_exception import NoValidConnectionsError
from paramiko.ssh_exception import SSHException
from pydantic import validate_call
from shepherd_core import Inventory
from shepherd_core import local_tz
from shepherd_core.data_models import ShpModel
from shepherd_core.data_models import Wrapper
from shepherd_core.data_models.base.shepherd import path_to_str
from shepherd_core.data_models.task import extract_tasks
from shepherd_core.data_models.task import prepare_task
from shepherd_core.data_models.testbed import Testbed
from shepherd_core.testbed_client import tb_client
from tqdm import tqdm
from typing_extensions import Self

from .logger import log


def _get_xdg_path(variable_name: str, default: str) -> Path:
    _value = os.environ.get(variable_name)
    if _value is None or _value == "":
        return Path("~").expanduser() / default
    return Path(_value)


path_xdg_config = _get_xdg_path("XDG_CONFIG_HOME", ".config/")


class Herd:
    path_default = PurePosixPath("/var/shepherd/recordings/")
    _remote_paths_allowed: AbstractSet[Path] = {
        path_default,
        PurePosixPath("/var/shepherd/"),
        PurePosixPath("/etc/shepherd/"),
        PurePosixPath("/tmp/"),  # noqa: S108
    }

    timestamp_diff_allowed = 10
    start_delay_s = 40

    def __init__(
        self,
        inventory: str | Path | None = None,
        limit: str | None = None,
        user: str | None = None,
        key_filepath: Path | None = None,
    ) -> None:
        limits_list: list[str] | None = None
        if isinstance(limit, str):
            limits_list = limit.split(",")
            limits_list = [_host for _host in limits_list if len(_host) >= 1]
        if isinstance(inventory, str) and Path(inventory).exists() and Path(inventory).is_file():
            inventory = Path(inventory)
        if isinstance(inventory, str):
            hostlist = inventory.split(",")
            hostlist = [_host for _host in hostlist if len(_host) >= 1]
            if limits_list is not None:
                hostlist = list(set(hostlist) & set(limits_list))
            hostnames = {hostname: hostname for hostname in hostlist}
        else:
            # look at all these directories for inventory-file
            inventories: list[Path] = [  # highest prio first
                Path().cwd() / "herd.yaml",
                Path().cwd() / "inventory/herd.yaml",
                Path("~").expanduser() / "herd.yaml",
                path_xdg_config / "shepherd/herd.yaml",
                Path("/etc/shepherd/herd.yaml"),
            ]
            if isinstance(inventory, Path):
                inventories = [inventory, *inventories]
            host_path = None
            for inventory in inventories:
                _path = Path(inventory)
                if _path.exists() and _path.is_file():
                    host_path = _path
                    break

            if host_path is None:
                raise FileNotFoundError(", ".join(_i.as_posix() for _i in inventories))

            with host_path.open(encoding="utf-8-sig") as stream:
                try:
                    inventory_data = yaml.safe_load(stream)
                except yaml.YAMLError as _xpt:
                    msg = (
                        f"Couldn't read inventory file {host_path.as_posix()}, "
                        f"please provide a valid one"
                    )
                    raise FileNotFoundError(msg) from _xpt
            log.info("Shepherd-Inventory = '%s'", host_path.as_posix())

            hostlist = []
            hostnames: dict[str, str] = {}
            for hostname, hostvars in inventory_data["sheep"]["hosts"].items():
                if isinstance(limits_list, list) and (hostname not in limits_list):
                    continue

                if "ansible_host" in hostvars:
                    hostlist.append(hostvars["ansible_host"])
                    hostnames[hostvars["ansible_host"]] = hostname
                else:
                    hostlist.append(hostname)
                    hostnames[hostname] = hostname

            if user is None:
                with contextlib.suppress(KeyError):
                    user = inventory_data["sheep"]["vars"]["ansible_user"]

        if user is None:
            raise ValueError("Provide user by command line or in inventory file")

        if len(hostlist) < 1 or len(hostnames) < 1:
            raise ValueError(
                "Provide remote hosts (either inventory empty or limit does not match)",
            )

        connect_kwargs: dict[str, str] = {}
        if key_filepath is not None:
            connect_kwargs["key_filename"] = str(key_filepath)

        self.group: Group = Group(
            *hostlist,
            user=user,
            connect_timeout=5,
            connect_kwargs=connect_kwargs,
        )
        self.hostnames: dict[str, str] = hostnames

        log.info("Herd consists of %d sheep", len(self.group))

    def __del__(self) -> None:
        # ... overcautious closing of connections
        if not hasattr(self, "group") or not isinstance(self.group, Group):
            return
        with contextlib.suppress(TypeError):
            for cnx in self.group:
                cnx.close()
                del cnx

    def __enter__(self) -> Self:
        self._open()
        if len(self.group) < 1:
            log.error("No remote sheep in current herd! Will run dry")
        return self

    def __exit__(
        self,
        typ: type[BaseException] | None = None,
        exc: BaseException | None = None,
        tb: TracebackType | None = None,
        extra_arg: int = 0,
    ) -> None:
        if not hasattr(self, "group") or not isinstance(self.group, Group):
            return
        with contextlib.suppress(TypeError):
            for cnx in self.group:
                cnx.close()

    def __getitem__(self, key: str) -> Any:
        if key in self.hostnames:
            return self.hostnames[key]
        raise KeyError

    def __repr__(self) -> dict:
        return self.hostnames

    @staticmethod
    def _thread_open(
        cnx: Connection,
    ) -> None:
        if cnx.is_connected:
            return
        try:
            cnx.open()
        except (NoValidConnectionsError, SSHException, TimeoutError):
            log.error(
                "[%s] failed to open connection -> will exclude node from inventory",
                cnx.host,  # = IP
            )
            cnx.close()

    def _open(self) -> None:
        """Open Connection on all Nodes."""
        threads = {}
        for cnx in self.group:
            _name = self.hostnames[cnx.host]
            threads[_name] = threading.Thread(target=self._thread_open, args=[cnx])
            threads[_name].start()
        for host, thread in threads.items():
            thread.join(timeout=10.0)
            if thread.is_alive():
                log.error(
                    "Connection.Open() did fail to finish on %s - will delete that thread",
                    host,
                )
            del thread  # ... overcautious
        self.group = [cnx for cnx in self.group if cnx.is_connected]

    @staticmethod
    def _thread_run(
        cnx: Connection,
        sudo: bool,  # noqa: FBT001
        cmd: str,
        results: dict[str, Result],
        hostname: str,
    ) -> None:
        if not cnx.is_connected:
            return
        try:
            if sudo:
                results[hostname] = cnx.sudo(cmd, warn=True, hide=True)
            else:
                results[hostname] = cnx.run(cmd, warn=True, hide=True)
        except (NoValidConnectionsError, SSHException, TimeoutError):
            log.error(
                "[%s] failed to run '%s' -> will exclude node from inventory",
                cnx.host,  # IP
                cmd,
            )
            cnx.close()

    @validate_call
    def run_cmd(
        self,
        cmd: str,
        exclusive_host: str | None = None,
        *,
        sudo: bool = False,
        verbose: bool = True,
    ) -> dict[str, Result]:
        """Run COMMAND on the shell -> Returns output-results.

        NOTE: in case of error on a node that corresponding dict value is unavailable
        """
        results: dict[str, Result] = {}
        threads = {}
        level = logging.INFO if verbose else logging.DEBUG
        log.log(level, "Sheep-CMD = %s", cmd)
        for cnx in self.group:
            _name = self.hostnames[cnx.host]
            if exclusive_host and _name != exclusive_host:
                continue
            threads[_name] = threading.Thread(
                target=self._thread_run,
                args=(cnx, sudo, cmd, results, _name),
            )
            threads[_name].start()
        for host, thread in tqdm(
            threads.items(), desc="  .. joining threads", unit="n", leave=False
        ):
            thread.join()  # timeout=10.0
            if thread.is_alive():
                log.error(
                    "Command.Run() did fail to finish on %s - will delete that thread",
                    host,
                )
            del thread  # ... overcautious
        if len(results) < 1:
            log.error("ZERO nodes answered - check your config")
        return results

    @staticmethod
    def print_output(
        replies: Mapping[str, Result],
        *,
        verbose: bool = False,
    ) -> None:
        """Log output-results of shell commands."""
        # sort dict by key first
        replies = dict(sorted(replies.items()))
        for hostname, reply in replies.items():
            if not verbose and reply.exited == 0:
                continue
            if len(reply.stdout) > 0:
                log.info("\n************** %s - stdout **************", hostname)
                log.info(reply.stdout)
            if len(reply.stderr) > 0:
                log.error("\n~~~~~~~~~~~~~~ %s - stderr ~~~~~~~~~~~~~~", hostname)
                log.error(reply.stderr)
            log.info("Exit-code of %s = %s", hostname, reply.exited)

    @staticmethod
    def _thread_put(
        cnx: Connection,
        src: Path | BytesIO,
        dst: PurePosixPath,
        force_overwrite: bool,  # noqa: FBT001
    ) -> None:
        if isinstance(src, BytesIO):
            filename = dst.name
        else:
            filename = src.name
            src = str(src)

        if not dst.suffix and not str(dst).endswith("/"):
            dst = PurePosixPath(str(dst) + "/")

        if not cnx.is_connected:
            return

        tmp_path = PurePosixPath("/tmp") / filename  # noqa: S108
        log.debug("temp-path for %s is %s", cnx.host, tmp_path)
        try:
            cnx.put(src, tmp_path.as_posix())
            xtr_arg = "-f" if force_overwrite else "-n"
            cnx.sudo(f"mv {xtr_arg} {tmp_path} {dst}", warn=True, hide=True)
        except (NoValidConnectionsError, SSHException, TimeoutError):
            log.error(
                "[%s] failed to put to '%s' -> will exclude node from inventory",
                cnx.host,
                dst.as_posix(),
            )
            cnx.close()

    def put_file(
        self,
        src: BytesIO | Path | str,
        dst: PurePosixPath | str,
        *,
        force_overwrite: bool = False,
    ) -> None:
        if isinstance(src, BytesIO):
            log.warning("BytesIO is buggy on some py-versions (only partial copy) -> use paths")
            src_path = src
        else:
            src_path = Path(src).absolute()
            if not src_path.exists():
                msg = f"Local source file '{src_path}' does not exist!"
                raise FileNotFoundError(msg)
            log.info("Local source path = %s", src_path)

        if dst is None:
            dst_path = self.path_default
            log.debug("Remote path not provided -> use default = %s", dst_path)
        else:
            dst_path = PurePosixPath(dst)
            dst_posix = dst_path.as_posix()
            is_allowed = False
            for path_allowed in self._remote_paths_allowed:
                if dst_posix.startswith(path_allowed.as_posix()):
                    is_allowed = True
            if not is_allowed:
                msg = f"provided path was forbidden ('{dst_posix}')"
                raise NameError(msg)

        threads = {}
        for cnx in self.group:
            _name = self.hostnames[cnx.host]
            threads[_name] = threading.Thread(
                target=self._thread_put,
                args=(cnx, src_path, dst_path, force_overwrite),
            )
            threads[_name].start()
        for host, thread in tqdm(
            threads.items(), desc="  .. joining threads", unit="n", leave=False
        ):
            thread.join()  # timeout=10.0
            if thread.is_alive():
                log.error(
                    "File.Put() did fail to finish on %s - will delete that thread",
                    host,
                )
            del thread  # ... overcautious

    @staticmethod
    def _thread_get(cnx: Connection, src: PurePosixPath, dst: Path) -> None:
        if not cnx.is_connected:
            return
        try:
            cnx.get(src.as_posix(), local=dst.as_posix())
        except (NoValidConnectionsError, SSHException, TimeoutError):
            log.error(
                "[%s] failed to get '%s' -> will exclude node from inventory",
                cnx.host,
                src.as_posix(),
            )
            cnx.close()

    @validate_call
    def get_file(
        self,
        src: PurePosixPath | str,
        dst_dir: Path | str,
        exclusive_host: str | None = None,
        *,
        timestamp: bool = False,
        separate: bool = False,
        delete_src: bool = False,
    ) -> bool:
        time_str = time.strftime("%Y_%m_%dT%H_%M_%S")
        xtra_ts = f"_{time_str}" if timestamp else ""
        failed_retrieval = False

        threads = {}
        dst_paths = {}

        # assemble file-names
        src_path: PurePosixPath = (
            PurePosixPath(src) if PurePosixPath(src).is_absolute() else self.path_default / src
        )

        for i, cnx in enumerate(self.group):
            hostname = self.hostnames[cnx.host]
            if separate:
                target_path = Path(dst_dir) / hostname
                xtra_node = ""
            else:
                target_path = Path(dst_dir)
                xtra_node = "" if hostname in src_path.stem else f"_{hostname}"

            dst_paths[i] = target_path / (src_path.stem + xtra_ts + xtra_node + src_path.suffix)

        # check if file is present
        replies = self.run_cmd(
            sudo=False, exclusive_host=exclusive_host, cmd=f"test -f {src_path}", verbose=False
        )

        # try to fetch data
        for i, cnx in enumerate(self.group):
            hostname = self.hostnames[cnx.host]
            if not isinstance(replies.get(hostname), Result):
                continue
            if abs(replies[hostname].exited) != 0:
                log.error(
                    "remote file '%s' does not exist on node %s",
                    src_path,
                    hostname,
                )
                failed_retrieval = True
                continue

            if not dst_paths[i].parent.exists():
                log.info("creating local dir of %s", dst_paths[i])
                dst_paths[i].parent.mkdir()

            log.debug(
                "retrieving remote src-file '%s' from %s to local dst '%s'",
                src_path,
                hostname,
                dst_paths[i],
            )

            threads[i] = threading.Thread(
                target=self._thread_get,
                args=(cnx, src_path, dst_paths[i]),
            )
            threads[i].start()
        log.debug("  .. threads started - will wait until finished")
        for i, cnx in enumerate(
            tqdm(self.group, desc="  .. joining threads", unit="n", leave=False)
        ):
            hostname = self.hostnames[cnx.host]
            if not isinstance(replies.get(hostname), Result):
                continue
            if replies[hostname].exited != 0:
                continue
            threads[i].join()  # timeout=10.0
            if threads[i].is_alive():
                log.error(
                    "Command.Run() did fail to finish on %s - will delete that thread",
                    hostname,
                )
            del threads[i]  # ... overcautious
            if delete_src:
                log.info(
                    "deleting %s from remote %s",
                    src_path,
                    hostname,
                )
                cnx.sudo(f"rm {src_path}", hide=True)

        del threads
        return failed_retrieval

    def find_consensus_time(self) -> tuple[datetime, float]:
        """Find a start time in the future when all nodes should start service.

        In order to run synchronously, all nodes should start at the same time.
        This is achieved by querying all nodes to check any large time offset,
        agreeing on a common time in the future and waiting for that time on each
        node.
        """
        # Get the current time on each target node
        replies = self.run_cmd(sudo=False, cmd="date --iso-8601=seconds", verbose=False)
        ts_nows = [datetime.fromisoformat(reply.stdout.rstrip()) for reply in replies.values()]
        if len(ts_nows) == 0:
            raise RuntimeError("No active hosts found to synchronize.")
        ts_max = max(ts_nows)
        ts_min = min(ts_nows)
        ts_diff = ts_max.timestamp() - ts_min.timestamp()
        # Check for excessive time difference among nodes
        if ts_diff > self.timestamp_diff_allowed:
            msg = f"Time difference between hosts greater {self.timestamp_diff_allowed} s"
            raise RuntimeError(msg)
        if ts_max.tzinfo is None:
            log.error("Provided time from host should have time-zone data!")
        # We need to estimate a future point in time such that all nodes are ready
        ts_start = ts_max + timedelta(seconds=self.start_delay_s)
        return ts_start, float(self.start_delay_s + ts_diff / 2)

    @validate_call
    def put_task(
        self,
        task: Path | ShpModel,
        remote_path: PurePosixPath | str = "/etc/shepherd/config.pickle",
    ) -> None:
        """Transfer shepherd tasks to the group of hosts / sheep.

        Rolls out a configuration file according to the given command and parameters
        service.

        """
        if remote_path.suffix.lower() != ".pickle":
            raise NameError("Remote path must point to '.pickle'")

        with TemporaryDirectory() as temp_dir:
            if isinstance(task, ShpModel):  # Model gets pickled to file
                task_dict = task.model_dump(exclude_unset=True)
                task_wrap = Wrapper(
                    datatype=type(task).__name__,
                    created=datetime.now(tz=local_tz()),
                    parameters=task_dict,
                )
                wrap_dict = path_to_str(task_wrap.model_dump(exclude_unset=True))
                task = Path(temp_dir) / "herd.pickle"
                # NOTE: preferred way is ByteIO/StringIO, but it is highly buggy
                with task.open("wb") as fd:
                    pickle.dump(wrap_dict, fd)

            elif isinstance(task, Path):  # file gets pickled if it is YAML (speedup sheep)
                if not task.is_file() or not task.exists():
                    raise FileNotFoundError("Task-Path must be an existing file")
                if task.is_file():
                    task_wrap = prepare_task(task)  # also functions as test
                    if task.suffix.lower() == ".yaml":  # repickle to file
                        wrap_dict = path_to_str(task_wrap.model_dump(exclude_unset=True))
                        task = Path(temp_dir) / "herd.pickle"
                        with task.open("wb") as fd:
                            pickle.dump(wrap_dict, fd)
            else:
                raise TypeError("Task must either be model or path to a model")

            if self.check_status(warn=True):
                raise RuntimeError("Shepherd still active!")
            if not isinstance(remote_path, PurePath):
                remote_path = PurePosixPath(remote_path)

            log.info(
                "Rolling out the config to '%s'",
                remote_path.as_posix(),
            )
            self.put_file(
                task,
                remote_path,
                force_overwrite=True,
            )

    @validate_call
    def check_status(self, *, warn: bool = False) -> bool:
        """Return true as long as one instance is still measuring.

        :param warn:
        :return: True is one node is still active
        """
        replies = self.run_cmd(sudo=True, cmd="systemctl status shepherd", verbose=False)
        active = False

        for cnx in self.group:
            hostname = self.hostnames[cnx.host]
            if not isinstance(replies.get(hostname), Result):
                continue
            if replies[hostname].exited != 3:
                active = True
                if warn:
                    log.warning(
                        "shepherd still active on %s",
                        hostname,
                    )
                else:
                    log.debug(
                        "shepherd still active on %s",
                        hostname,
                    )
                    # shepherd-herd -v shell-cmd -s "systemctl status shepherd"
        return active

    def get_last_usage(self) -> timedelta | None:
        """Gives time-delta of last testbed usage."""
        replies1 = self.run_cmd(sudo=True, cmd="tail -n 1 /var/shepherd/log.csv", verbose=False)
        replies2 = self.run_cmd(sudo=False, cmd="date --iso-8601=seconds", verbose=False)
        deltas = []
        for cnx in self.group:
            hostname = self.hostnames[cnx.host]
            if not isinstance(replies1.get(hostname), Result):
                continue
            if not isinstance(replies2.get(hostname), Result):
                continue
            if replies1[hostname].exited == 0:
                ts_end = datetime.fromisoformat(replies1[hostname].stdout.split(",")[1].strip())
                ts_now = datetime.fromisoformat(replies2[hostname].stdout.strip())
                deltas.append(ts_now - ts_end)
        if len(deltas) == 0:
            return None
        return min(deltas)

    def start_measurement(self) -> int:
        """Start shepherd service on the group of hosts."""
        if self.check_status(warn=True):
            log.info("-> won't start while shepherd-instances are active")
            return 1

        replies = self.run_cmd(sudo=True, cmd="systemctl start shepherd")
        self.print_output(replies)
        return max([0] + [abs(reply.exited) for reply in replies.values()])

    def stop_measurement(self) -> int:
        log.debug("Shepherd-nodes affected: %s", self.hostnames.values())
        replies = self.run_cmd(sudo=True, cmd="systemctl stop shepherd")
        exit_code = max([0] + [abs(reply.exited) for reply in replies.values()])
        log.info("Shepherd was forcefully stopped")
        if exit_code > 0:
            log.debug("-> max exit-code = %d", exit_code)
        return exit_code

    @validate_call
    def poweroff(self, *, restart: bool) -> int:
        log.debug("Shepherd-nodes affected: %s", self.hostnames.values())
        if restart:
            replies = self.run_cmd(sudo=True, cmd="reboot")
            log.info("Command for rebooting nodes was issued")
        else:
            replies = self.run_cmd(sudo=True, cmd="poweroff")
            log.info("Command for powering off nodes was issued")
        return max([0] + [abs(reply.exited) for reply in replies.values()])

    @validate_call
    def await_stop(self, timeout: int = 30) -> bool:
        ts_end = time.time() + timeout
        while self.check_status():
            if time.time() > ts_end:
                return self.check_status(warn=True)
            time.sleep(1)
        return False

    @validate_call
    def inventorize(self, output_path: Path) -> bool:
        """Collect information about the hosts, including the herd-server."""
        if output_path.is_file():
            msg = f"Inventorize needs a dir, not a file '{output_path.as_posix()}'"
            raise ValueError(msg)
        file_path = PurePosixPath("/var/shepherd/inventory.yaml")
        self.run_cmd(
            sudo=True,
            cmd=f"shepherd-sheep inventorize --output-path {file_path.as_posix()}",
        )
        server_inv = Inventory.collect()
        output_path = Path(output_path)
        server_inv.to_file(
            path=Path(output_path) / "inventory_server.yaml",
            minimal=True,
        )
        return self.get_file(
            file_path,
            output_path,
            timestamp=False,
            separate=False,
            delete_src=True,
        )
        # TODO: best case - add all to one file or a new inventories-model?

    def resync(self) -> int:
        """Get current time via ntp and restart PTP on each sheep."""
        commands = [
            "systemctl stop phc2sys@eth0",
            "systemctl stop ptp4l@eth0",
            "ntpdate -s time.nist.gov",
            "systemctl start phc2sys@eth0",
            "systemctl start ptp4l@eth0",
            "shepherd-sheep fix",  # restarts kernel module
        ]
        exit_code = 0
        for command in commands:
            ret = self.run_cmd(sudo=True, cmd=command)
            self.print_output(ret, verbose=True)
            exit_code = max([exit_code] + [abs(reply.exited) for reply in ret.values()])

        # Get the current time on each target node
        replies_date = self.run_cmd(sudo=False, cmd="date --iso-8601=seconds", verbose=False)
        self.print_output(replies_date, verbose=True)
        # calc diff
        ts_nows = [datetime.fromisoformat(reply.stdout.rstrip()) for reply in replies_date.values()]
        if len(ts_nows) == 0:
            raise RuntimeError("No active hosts found to synchronize.")
        ts_max = max(ts_nows)
        ts_min = min(ts_nows)
        ts_diff = ts_max.timestamp() - ts_min.timestamp()
        if ts_diff > 5:
            log.error("Timediff after resync is too large (%d s)", ts_diff)
            return 1
        log.info("Timediff is OK (%d s)", ts_diff)
        return exit_code

    @validate_call
    def run_task(
        self, config: Path | ShpModel, *, attach: bool = False, quiet: bool = False
    ) -> int:
        if attach:
            remote_path = PurePosixPath("/etc/shepherd/config_for_herd.pickle")
            self.put_task(config, remote_path)
            command = f"shepherd-sheep --verbose run {remote_path.as_posix()}"
            replies = self.run_cmd(sudo=True, cmd=command)
            exit_code = max([0] + [abs(reply.exited) for reply in replies.values()])
            if exit_code:
                log.error("Running Task failed - will exit now!")
            if not quiet:
                self.print_output(replies, verbose=True)
        else:
            remote_path = PurePosixPath("/etc/shepherd/config.pickle")
            self.put_task(config, remote_path)
            exit_code = self.start_measurement()
            log.info("Shepherd started.")
            if exit_code > 0:
                log.debug("-> max exit-code = %d", exit_code)
        return exit_code

    @validate_call
    def get_task_files(
        self,
        config: Path | ShpModel,
        dst_dir: Path | str,
        *,
        separate: bool = False,
        delete_src: bool = False,
    ) -> bool:
        tbed_id = tb_client.query_ids("Testbed")[0]
        tbed_di = tb_client.query_item("Testbed", tbed_id)
        tbed = Testbed(**tbed_di)
        if tbed.shared_storage:
            log.info("Data should be locally at: %s", {tbed.data_on_server})

        wrap = prepare_task(config)
        tasks = extract_tasks(wrap, no_task_sets=False)
        failed = False
        for task in tasks:
            if hasattr(task, "output_path"):
                log.info("General remote path is: %s", task.output_path)
                failed |= self.get_file(
                    task.output_path,
                    dst_dir,
                    separate=separate,
                    delete_src=delete_src,
                )
            elif hasattr(task, "get_output_paths"):
                for host, path in task.get_output_paths().items():
                    log.info("Remote path of '%s' is: %s", host, path)
                    failed |= self.get_file(
                        path,
                        dst_dir,
                        exclusive_host=host,
                        separate=separate,
                        delete_src=delete_src,
                    )
        return failed

    def alive(self) -> bool:
        """Check if all remote hosts are present & responding.

        - Group is list of hosts with live connection,
        - hostnames contains all hosts in inventory
        """
        return len(self.group) >= len(self.hostnames)



import logging

from shepherd_core.logger import get_verbose_level
from shepherd_core.logger import set_log_verbose_level

log = logging.getLogger("shepherd-herd")
log.addHandler(logging.NullHandler())
set_log_verbose_level(log, 2)
# Note: defined here to avoid circular import
# TODO: add queue and also save log to file


def get_verbosity() -> bool:
    return get_verbose_level() >= 3


def activate_verbosity() -> None:
    set_log_verbose_level(log, 3)



from pathlib import Path

import pytest
from click.testing import CliRunner
from shepherd_herd.herd_cli import cli

# NOTE: (almost) direct copy between shepherd-herd & python-package
# differences: import _herd, .mark.hardware, shepherd_up / stopped_herd


@pytest.fixture
def fw_example() -> Path:
    here = Path(__file__).absolute()
    name = "firmware_nrf52_powered.hex"
    return here.parent / name


@pytest.fixture
def fw_empty(tmp_path: Path) -> Path:
    store_path = tmp_path / "firmware_null.hex"
    with store_path.open("w", encoding="utf-8-sig") as f:
        f.write("")
    return store_path


@pytest.mark.timeout(80)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_minimal(cli_runner: CliRunner, fw_example: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--simulate",
            fw_example.as_posix(),
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(80)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_swd_explicit(cli_runner: CliRunner, fw_example: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--target-port",
            "A",
            "--voltage",
            "2.0",
            "--datarate",
            "600000",
            "--mcu-type",
            "nrf52",
            "--mcu-port",
            "1",
            "--simulate",
            fw_example.as_posix(),
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(80)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_swd_explicit_short(
    cli_runner: CliRunner,
    fw_example: Path,
) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "-p",
            "A",
            "-v",
            "2.0",
            "-d",
            "600000",
            "-t",
            "nrf52",
            "-m",
            "1",
            "--simulate",
            fw_example.as_posix(),
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(80)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_sbw_explicit(cli_runner: CliRunner, fw_example: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--target-port",
            "B",
            "--voltage",
            "1.5",
            "--datarate",
            "300000",
            "--mcu-type",
            "msp430",
            "--mcu-port",
            "2",
            "--simulate",
            fw_example.as_posix(),
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(60)  # TODO: fails in big runs - so add alive-checker here
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_file_defective_a(cli_runner: CliRunner, fw_empty: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--simulate",
            fw_empty.as_posix(),
        ],
    )
    assert res.exit_code != 0


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_file_defective_b(cli_runner: CliRunner, tmp_path: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--simulate",
            tmp_path.as_posix(),  # Directory
        ],
    )
    assert res.exit_code != 0


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_file_defective_c(cli_runner: CliRunner, tmp_path: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--simulate",
            str(tmp_path / "file_abc.bin"),  # non_existing file
        ],
    )
    assert res.exit_code != 0


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_datarate_invalid_a(
    cli_runner: CliRunner,
    fw_example: Path,
) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--datarate",
            "2000000",  # too fast
            "--simulate",
            fw_example.as_posix(),
        ],
    )
    assert res.exit_code != 0


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_datarate_invalid_b(
    cli_runner: CliRunner,
    fw_example: Path,
) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--datarate",
            "0",  # impossible
            "--simulate",
            fw_example.as_posix(),
        ],
    )
    assert res.exit_code != 0


@pytest.mark.timeout(60)
@pytest.mark.usefixtures("_herd_alive")
def test_cli_program_target_invalid(cli_runner: CliRunner, fw_example: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "program",
            "--mcu-type",
            "arduino",
            "--simulate",
            fw_example.as_posix(),
        ],
    )
    assert res.exit_code != 0


# not testable ATM (through CLI)
#   - fail pins 3x (pin_num is identical)
#   - fail wrong target (internally, fail in kModule)
#   - datasize > mem_size



from pathlib import Path

import pytest
from click.testing import CliRunner
from shepherd_herd.herd_cli import cli

from .conftest import extract_first_sheep
from .conftest import generate_h5_file


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_run_standard(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "shell",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_run_extra(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "shell",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_run_fail(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "shell-command",
            "date",
        ],
    )
    assert res.exit_code != 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_run_sudo(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "shell",
            "-s",
            "echo 'it's me: $USER",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_run_sudo_long(cli_runner: CliRunner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "shell",
            "--sudo",
            "echo 'it's me: $USER",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_provide_inventory(cli_runner: CliRunner, local_herd: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "-i",
            local_herd.as_posix(),
            "shell",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_provide_inventory_long(cli_runner: CliRunner, local_herd: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "--inventory",
            local_herd.as_posix(),
            "--verbose",
            "shell",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_provide_limit(cli_runner: CliRunner, local_herd: Path) -> None:
    sheep = extract_first_sheep(local_herd)
    res = cli_runner.invoke(
        cli,
        [
            "-i",
            local_herd.as_posix(),
            "-l",
            f"{sheep},",
            "-v",
            "shell",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("_herd_alive")
def test_provide_limit_long(cli_runner: CliRunner, local_herd: Path) -> None:
    sheep = extract_first_sheep(local_herd)
    res = cli_runner.invoke(
        cli,
        [
            "-i",
            local_herd.as_posix(),
            "--limit",
            f"{sheep},",
            "-v",
            "shell",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_provide_limit_fail(cli_runner: CliRunner, local_herd: Path) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "-i",
            local_herd.as_posix(),
            "-l",
            "MrMeeseeks,",
            "shell",
            "date",
        ],
    )
    assert res.exit_code != 0


@pytest.mark.usefixtures("_herd_alive")
def test_distribute_retrieve_std(cli_runner: CliRunner, tmp_path: Path) -> None:
    test_file = generate_h5_file(tmp_path, "pytest_deploy.h5")
    elem_count1 = len(list(tmp_path.iterdir()))
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "distribute",
            test_file.as_posix(),
        ],
    )
    assert res.exit_code == 0
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "retrieve",
            "-f",
            "-t",
            "-d",
            test_file.name,
            tmp_path.as_posix(),
        ],
    )
    assert res.exit_code == 0
    elem_count2 = len(list(tmp_path.iterdir()))
    # file got deleted in prev retrieve, so fail now
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "retrieve",
            "-s",
            test_file.name,
            tmp_path.as_posix(),
        ],
    )
    assert res.exit_code != 0
    elem_count3 = len(list(tmp_path.iterdir()))
    assert elem_count1 < elem_count2
    assert elem_count2 == elem_count3


@pytest.mark.usefixtures("_herd_alive")
def test_distribute_retrieve_etc(cli_runner: CliRunner, tmp_path: Path) -> None:
    test_file = generate_h5_file(tmp_path, "pytest_deploy.h5")
    elem_count1 = len(list(tmp_path.iterdir()))
    dir_remote = "/etc/shepherd/"
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "distribute",
            "--remote-path",
            dir_remote,
            test_file.as_posix(),
        ],
    )
    assert res.exit_code == 0
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "retrieve",
            "--force-stop",
            "--separate",
            "--delete",
            dir_remote + test_file.name,
            tmp_path.as_posix(),
        ],
    )
    assert res.exit_code == 0
    elem_count2 = len(list(tmp_path.iterdir()))
    # file got deleted in prev retrieve, so fail now
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "retrieve",
            "--timestamp",
            dir_remote + test_file.name,
            tmp_path.as_posix(),
        ],
    )
    assert res.exit_code != 0
    elem_count3 = len(list(tmp_path.iterdir()))
    assert elem_count1 < elem_count2
    assert elem_count2 == elem_count3


@pytest.mark.usefixtures("_herd_alive")
def test_distribute_retrieve_var(cli_runner: CliRunner, tmp_path: Path) -> None:
    test_file = generate_h5_file(tmp_path, "pytest_deploy.h5")
    elem_count1 = len(list(tmp_path.iterdir()))
    dir_remote = "/var/shepherd/"
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "distribute",
            "-r",
            dir_remote,
            test_file.as_posix(),
        ],
    )
    assert res.exit_code == 0
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "retrieve",
            "--force-stop",
            "--separate",
            "--delete",
            dir_remote + test_file.name,
            tmp_path.as_posix(),
        ],
    )
    assert res.exit_code == 0
    elem_count2 = len(list(tmp_path.iterdir()))
    # file got deleted in prev retrieve, so fail now
    res = cli_runner.invoke(
        cli,
        [
            "-v",
            "retrieve",
            "--timestamp",
            dir_remote + test_file.name,
            tmp_path.as_posix(),
        ],
    )
    assert res.exit_code != 0
    elem_count3 = len(list(tmp_path.iterdir()))
    assert elem_count1 < elem_count2
    assert elem_count2 == elem_count3


# TODO: test providing user and key filename
# TODO: test poweroff (reboot)
# TODO: test sudo
